@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2005 to 2007 by Jonathan Duddington                     *
+ *   Copyright (C) 2005 to 2010 by Jonathan Duddington                     *
  *   email: jonsd@users.sourceforge.net                                    *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
@@ -54,6 +54,7 @@ static int  last_wcmdq;
 static int  pitch_length;
 static int  amp_length;
 static int  modn_flags;
+static int  fmt_amplitude=0;
 
 static int  syllable_start;
 static int  syllable_end;
@@ -135,7 +136,7 @@ static void EndPitch(int voice_break)
 		syllable_centre = -1;
 		memset(vowel_transition,0,sizeof(vowel_transition));
 	}
-}  // end of Synthesize::EndPitch
+}  // end of EndPitch
 
 
 
@@ -152,7 +153,7 @@ static void DoAmplitude(int amp, unsigned char *amp_env)
 	q[2] = (long)amp_env;
 	q[3] = amp;
 	WcmdqInc();
-}  // end of Synthesize::DoAmplitude
+}  // end of DoAmplitude
 
 
 
@@ -162,11 +163,11 @@ static void DoPitch(unsigned char *env, int pitch1, int pitch2)
 
 	EndPitch(0);
 
-	if(pitch1 == 1024)
+	if(pitch1 == 255)
 	{
 		// pitch was not set
-		pitch1 = 24;
-		pitch2 = 33;
+		pitch1 = 55;
+		pitch2 = 76;
 		env = envelope_data[PITCHfall];
 	}
 	last_pitch_cmd = wcmdq_tail;
@@ -181,7 +182,7 @@ static void DoPitch(unsigned char *env, int pitch1, int pitch2)
 	q[2] = (long)env;
 	q[3] = (pitch1 << 16) + pitch2;
 	WcmdqInc();
-}  //  end of Synthesize::DoPitch
+}  //  end of DoPitch
 
 
 
@@ -190,9 +191,14 @@ int PauseLength(int pause, int control)
 	int len;
 
 	if(control == 0)
-		len = (pause * speed.speed_factor1)/256;
+	{
+		if(pause >= 200)
+			len = (pause * speed.clause_pause_factor)/256;
+		else
+			len = (pause * speed.pause_factor)/256;
+	}
 	else
-		len = (pause * speed.speed_factor2)/256;
+		len = (pause * speed.wav_factor)/256;
 
 	if(len < 5) len = 5;      // mS, limit the amount to which pauses can be shortened
 	return(len);
@@ -204,110 +210,223 @@ static void DoPause(int length, int control)
 // control = 1, less shortening at fast speeds
 	int len;
 
-	len = PauseLength(length, control);
+	if(length == 0)
+		len = 0;
+	else
+	{
+		len = PauseLength(length, control);
 
-	len = (len * samplerate) / 1000;  // convert from mS to number of samples
+		len = (len * samplerate) / 1000;  // convert from mS to number of samples
+	}
 
 	EndPitch(1);
 	wcmdq[wcmdq_tail][0] = WCMD_PAUSE;
 	wcmdq[wcmdq_tail][1] = len;
 	WcmdqInc();
 	last_frame = NULL;
-}  // end of Synthesize::DoPause
+
+	if(fmt_amplitude != 0)
+	{
+		wcmdq[wcmdq_tail][0] = WCMD_FMT_AMPLITUDE;
+		wcmdq[wcmdq_tail][1] = fmt_amplitude = 0;
+		WcmdqInc();
+	}
+}  // end of DoPause
 
 
 extern int seq_len_adjust;   // temporary fix to advance the start point for playing the wav sample
 
 
-static int DoSample2(int index, int which, int length_mod, int amp)
-{//================================================================
+static int DoSample2(int index, int which, int std_length, int control, int length_mod, int amp)
+{//=============================================================================================
 	int length;
-	int length1;
-	int format;
+	int wav_length;
+	int wav_scale;
 	int min_length;
-	int start=0;
+	int x;
+	int len4;
 	long *q;
 	unsigned char *p;
 
 	index = index & 0x7fffff;
 	p = &wavefile_data[index];
-	format = p[2];
-	length1 = (p[1] * 256);
-	length1 += p[0];    //  length in bytes
+	wav_scale = p[2];
+	wav_length = (p[1] * 256);
+	wav_length += p[0];    //  length in bytes
 
-	if(seq_len_adjust > 0)
+	if(wav_length == 0)
+		return(0);
+
+	min_length = speed.min_sample_len;
+
+	if(wav_scale==0)
+		min_length *= 2;  // 16 bit samples
+	else
 	{
-		start = (seq_len_adjust * samplerate)/1000;
-		if(format == 0)
-			start *= 2;
-		length1 -= start;
-		index += start;
+		// increase consonant amplitude at high speeds, depending on the peak consonant amplitude
+//		x = ((35 - wav_scale) * speed.loud_consonants);
+//		if(x < 0) x = 0;
+//		wav_scale = (wav_scale * (x+256))/256;
 	}
 
+	if(std_length > 0)
+	{
+		std_length = (std_length * samplerate)/1000;
+		if(wav_scale == 0)
+			std_length *= 2;
+
+		x = (min_length * std_length)/wav_length;
+		if(x > min_length)
+			min_length = x;
+	}
+	else
+	{
+		// no length specified, use the length of the stored sound
+		std_length = wav_length;
+	}
 
 	if(length_mod > 0)
-		length = (length1 * length_mod) / 256;
-	else
-		length = length1;
+	{
+		std_length = (std_length * length_mod)/256;
+	}
 
+	length = (std_length * speed.wav_factor)/256;
 
-	length = (length * speed.speed_factor2)/256;
-	min_length = speed.min_sample_len;
-	if(format==0)
-		min_length *= 2;
+	if(control & pd_DONTLENGTHEN)
+	{
+		// this option is used for Stops, with short noise bursts.
+		// Don't change their length much.
+		if(length > std_length)
+		{
+			// don't let length exceed std_length
+			length = std_length;
+		}
+		else
+		{
+			// reduce the reduction in length
+//			length = (length + std_length)/2;
+		}
+	}
 
 	if(length < min_length)
 		length = min_length;
 
-	if(length > length1)
-		length = length1;  // don't exceed wavefile length
 
-	if(format==0)
-		length /= 2;     // 2 byte samples
+	if(wav_scale == 0)
+	{
+		// 16 bit samples
+		length /= 2;
+		wav_length /= 2;
+	}
 
+	if(amp < 0)
+		return(length);
+
+	len4 = wav_length / 4;
 
 	index += 4;
 
-	if(amp >= 0)
+	if(which & 0x100)
 	{
+		// mix this with synthesised wave
 		last_wcmdq = wcmdq_tail;
 		q = wcmdq[wcmdq_tail];
-		if(which & 0x100)
-			q[0] = WCMD_WAVE2;    // mix this with synthesised wave
-		else
-			q[0] = WCMD_WAVE;
-		q[1] = length;   // length in samples
+		q[0] = WCMD_WAVE2;
+		q[1] = length | (wav_length << 16);   // length in samples
 		q[2] = long(&wavefile_data[index]);
-		q[3] = format + (amp << 8);
+		q[3] = wav_scale + (amp << 8);
+		WcmdqInc();
+		return(length);
+	}
+
+	if(length > wav_length)
+	{
+		x = len4*3;
+		length -= x;
+	}
+	else
+	{
+		x = length;
+		length = 0;
+	}
+
+	last_wcmdq = wcmdq_tail;
+	q = wcmdq[wcmdq_tail];
+	q[0] = WCMD_WAVE;
+	q[1] = x;   // length in samples
+	q[2] = long(&wavefile_data[index]);
+	q[3] = wav_scale + (amp << 8);
+	WcmdqInc();
+
+
+	while(length > len4*3)
+	{
+		x = len4;
+		if(wav_scale == 0)
+			x *= 2;
+
+		last_wcmdq = wcmdq_tail;
+		q = wcmdq[wcmdq_tail];
+		q[0] = WCMD_WAVE;
+		q[1] = len4*2;   // length in samples
+		q[2] = long(&wavefile_data[index+x]);
+		q[3] = wav_scale + (amp << 8);
+		WcmdqInc();
+
+		length -= len4*2;
+	}
+
+	if(length > 0)
+	{
+		x = wav_length - length;
+		if(wav_scale == 0)
+			x *= 2;
+		last_wcmdq = wcmdq_tail;
+		q = wcmdq[wcmdq_tail];
+		q[0] = WCMD_WAVE;
+		q[1] = length;   // length in samples
+		q[2] = long(&wavefile_data[index+x]);
+		q[3] = wav_scale + (amp << 8);
 		WcmdqInc();
 	}
+
 	return(length);
-}  // end of Synthesize::DoSample2
+}  // end of DoSample2
 
 
-int DoSample(PHONEME_TAB *ph1, PHONEME_TAB *ph2, int which, int length_mod, int amp)
-{//====================== ==========================================================
-	int index;
-	int match_level;
+
+int DoSample3(PHONEME_DATA *phdata, int length_mod, int amp)
+{//=========================================================
 	int amp2;
-	int result;
-
+	int len;
 	EndPitch(1);
-	index = LookupSound(ph1,ph2,which & 0xff,&match_level,0);
-	if((index & 0x800000) == 0)
-		return(0);             // not wavefile data
-
-	amp2 = wavefile_amp;
-	if(amp != 0)
-		amp2 = (amp * wavefile_amp)/20;
 
 	if(amp == -1)
+	{
+		// just get the length, don't produce sound
 		amp2 = amp;
+	}
+	else
+	{
+		amp2 = phdata->sound_param[pd_WAV];
+		if(amp2 == 0)
+			amp2 = 100;
+		amp2 = (amp2 * 32)/100;
+	}
 
-	result = DoSample2(index,which,length_mod,amp2);
+	seq_len_adjust=0;
+
+	if(phdata->sound_addr[pd_WAV] == 0)
+	{
+		len = 0;
+	}
+	else
+	{
+		len = DoSample2(phdata->sound_addr[pd_WAV], 2, phdata->pd_param[pd_LENGTHMOD]*2, phdata->pd_control, length_mod, amp2);
+	}
 	last_frame = NULL;
-	return(result);
-}  // end of Synthesize::DoSample
+	return(len);
+}  // end of DoSample3
 
 
 
@@ -535,7 +654,7 @@ static short vcolouring[N_VCOLOUR][5] = {
 //	fprintf(stderr,"FMT%d %3s  %3d-%3d f1=%d  f2=%4d %4d %4d  f3=%4d %3d\n",
 //		which,WordToString(other_ph->mnemonic),len,rms,f1,f2,f2_min,f2_max,f3_adj,f3_amp);
 
-	if(other_ph->mnemonic == '?')
+	if((other_ph != NULL) && (other_ph->mnemonic == '?'))
 		flags |= 8;
 
 	if(which == 1)
@@ -546,14 +665,15 @@ static short vcolouring[N_VCOLOUR][5] = {
 		seq[0].length = VOWEL_FRONT_LENGTH;
 		if(len > 0)
 			seq[0].length = len;
-		seq[0].frflags |= FRFLAG_LEN_MOD;              // reduce length modification
-		fr->frflags |= FRFLAG_LEN_MOD;
+		seq[0].frflags |= FRFLAG_LEN_MOD2;              // reduce length modification
+		fr->frflags |= FRFLAG_LEN_MOD2;
 
 		next_rms = seq[1].frame->rms;
 
 if(voice->klattv[0])
 {
-	fr->klattp[KLATT_AV] = 53;   // reduce the amplituide of the start of a vowel
+//	fr->klattp[KLATT_AV] = 53;   // reduce the amplituide of the start of a vowel
+   fr->klattp[KLATT_AV] = seq[1].frame->klattp[KLATT_AV] - 4;
 }
 		if(f2 != 0)
 		{
@@ -860,10 +980,10 @@ static void StartSyllable(void)
 }
 
 
-int DoSpect(PHONEME_TAB *this_ph, PHONEME_TAB *prev_ph, PHONEME_TAB *next_ph,
-		int which, PHONEME_LIST *plist, int modulation)
-{//===================================================================================
-	// which  1  start of phoneme,   2 body and end
+
+int DoSpect2(PHONEME_TAB *this_ph, int which, FMT_PARAMS *fmt_params,  PHONEME_LIST *plist, int modulation)
+{//========================================================================================================
+	// which:  0 not a vowel, 1  start of vowel,   2 body and end of vowel
 	// length_mod: 256 = 100%
 	// modulation: -1 = don't write to wcmdq
 
@@ -876,23 +996,33 @@ int DoSpect(PHONEME_TAB *this_ph, PHONEME_TAB *prev_ph, PHONEME_TAB *next_ph,
 	int  ix;
 	long *q;
 	int  len;
-	int  match_level;
 	int  frame_length;
-	int  frame1_length;
-	int  frame2_length;
 	int  length_factor;
 	int  length_mod;
+	int  length_sum;
+	int  length_min;
 	int  total_len = 0;
 	static int wave_flag = 0;
 	int wcmd_spect = WCMD_SPECT;
+	int frame_lengths[N_SEQ_FRAMES];
+
+	if(fmt_params->fmt_addr == 0)
+		return(0);
 
 	length_mod = plist->length;
 	if(length_mod==0) length_mod=256;
 
+	length_min = (samplerate/70);  // greater than one cycle at low pitch (Hz)
+	if(which==2)
+	{
+		if((translator->langopts.param[LOPT_LONG_VOWEL_THRESHOLD] > 0) && ((this_ph->std_length >= translator->langopts.param[LOPT_LONG_VOWEL_THRESHOLD]) || (plist->synthflags & SFLAG_LENGTHEN) || (this_ph->phflags & phLONG)))
+			length_min *= 2;    // ensure long vowels are longer
+	}
+
 if(which==1)
 {
 	// limit the shortening of sonorants before shortened (eg. unstressed vowels)
-	if((this_ph->type==phLIQUID) || (prev_ph->type==phLIQUID) || (prev_ph->type==phNASAL))
+	if((this_ph->type==phLIQUID) || (plist[-1].type==phLIQUID) || (plist[-1].type==phNASAL))
 	{
 		if(length_mod < (len = translator->langopts.param[LOPT_SONORANT_MIN]))
 		{
@@ -902,15 +1032,30 @@ if(which==1)
 }
 
 	modn_flags = 0;
-	frames = LookupSpect(this_ph,prev_ph,next_ph,which,&match_level,&n_frames, plist);
+	frames = LookupSpect(this_ph, which, fmt_params, &n_frames, plist);
 	if(frames == NULL)
 		return(0);   // not found
 
+	if(fmt_params->fmt_amp != fmt_amplitude)
+	{
+		// an amplitude adjustment is specified for this sequence
+		q = wcmdq[wcmdq_tail];
+		q[0] = WCMD_FMT_AMPLITUDE;
+		q[1] = fmt_amplitude = fmt_params->fmt_amp;
+		WcmdqInc();
+	}
+
 	frame1 = frames[0].frame;
-	frame1_length = frames[0].length;
 	if(voice->klattv[0])
 		wcmd_spect = WCMD_KLATT;
 
+	wavefile_ix = fmt_params->wav_addr;
+
+	if(fmt_params->wav_amp == 0)
+		wavefile_amp = 32;
+	else
+		wavefile_amp = (fmt_params->wav_amp * 32)/100;
+
 	if(wavefile_ix == 0)
 	{
 		if(wave_flag)
@@ -960,29 +1105,50 @@ if(which==1)
 		syllable_centre = wcmdq_tail;
 	}
 
-	frame_length = frame1_length;
+	length_sum = 0;
+	for(frameix=1; frameix < n_frames; frameix++)
+	{
+		length_factor = length_mod;
+		if(frames[frameix-1].frflags & FRFLAG_LEN_MOD)     // reduce effect of length mod
+		{
+			length_factor = (length_mod*(256-speed.lenmod_factor) + 256*speed.lenmod_factor)/256;
+		}
+		else
+		if(frames[frameix-1].frflags & FRFLAG_LEN_MOD2)     // reduce effect of length mod, used for the start of a vowel
+		{
+			length_factor = (length_mod*(256-speed.lenmod2_factor) + 256*speed.lenmod2_factor)/256;
+		}
+
+		frame_length = frames[frameix-1].length;
+		len = (frame_length * samplerate)/1000;
+		len = (len * length_factor)/256;
+		length_sum += len;
+		frame_lengths[frameix] = len;
+	}
+
+	if((length_sum > 0) && (length_sum < length_min))
+	{
+		// lengthen, so that the sequence is greater than one cycle at low pitch
+		for(frameix=1; frameix < n_frames; frameix++)
+		{
+			frame_lengths[frameix] = (frame_lengths[frameix] * length_min) / length_sum;
+		}
+	}
+
 	for(frameix=1; frameix<n_frames; frameix++)
 	{
 		frame2 = frames[frameix].frame;
-		frame2_length = frames[frameix].length;
 
-		if((wavefile_ix != 0) && ((frame1->frflags & FRFLAG_DEFER_WAV)==0))
+		if((fmt_params->wav_addr != 0) && ((frame1->frflags & FRFLAG_DEFER_WAV)==0))
 		{
 			// there is a wave file to play along with this synthesis
 			seq_len_adjust = 0;
-			DoSample2(wavefile_ix,which+0x100,0,wavefile_amp);
+			DoSample2(fmt_params->wav_addr, which+0x100, 0, fmt_params->fmt_control, 0, wavefile_amp);
 			wave_flag = 1;
 			wavefile_ix = 0;
+			fmt_params->wav_addr = 0;
 		}
 
-		length_factor = length_mod;
-		if(frame1->frflags & FRFLAG_LEN_MOD)     // reduce effect of length mod
-		{
-			length_factor = (length_mod*(256-speed.speed_factor3) + 256*speed.speed_factor3)/256;
-		}
-		len = (frame_length * samplerate)/1000;
-		len = (len * length_factor)/256;
-
 		if(modulation >= 0)
 		{
 			if(frame1->frflags & FRFLAG_MODULATE)
@@ -993,13 +1159,13 @@ if(which==1)
 				modulation |= modn_flags;   // before or after a glottal stop
 		}
 
+		len = frame_lengths[frameix];
 		pitch_length += len;
 		amp_length += len;
 
-		if(frame_length < 2)
+		if(len == 0)
 		{
 			last_frame = NULL;
-			frame_length = frame2_length;
 			frame1 = frame2;
 		}
 		else
@@ -1017,16 +1183,27 @@ if(which==1)
 				WcmdqInc();
 			}
 			last_frame = frame1 = frame2;
-			frame_length = frame2_length;
 			total_len += len;
 		}
 	}
+
+	if((which != 1) && (fmt_amplitude != 0))
+	{
+		q = wcmdq[wcmdq_tail];
+		q[0] = WCMD_FMT_AMPLITUDE;
+		q[1] = fmt_amplitude = 0;
+		WcmdqInc();
+	}
+
+
 	return(total_len);
-}  // end of Synthesize::DoSpect
+}  // end of DoSpect
 
 
-static void DoMarker(int type, int char_posn, int length, int value)
-{//=================================================================
+
+
+void DoMarker(int type, int char_posn, int length, int value)
+{//==========================================================
 // This could be used to return an index to the word currently being spoken
 // Type 1=word, 2=sentence, 3=named marker, 4=play audio, 5=end
 	wcmdq[wcmdq_tail][0] = WCMD_MARKER;
@@ -1035,7 +1212,7 @@ static void DoMarker(int type, int char_posn, int length, int value)
 	wcmdq[wcmdq_tail][3] = value;
 	WcmdqInc();
 
-}  // end of Synthesize::DoMarker
+}  // end of DoMarker
 
 
 void DoVoiceChange(voice_t *v)
@@ -1051,18 +1228,23 @@ void DoVoiceChange(voice_t *v)
 }
 
 
-static void DoEmbedded(int &embix, int sourceix)
-{//=============================================
+void DoEmbedded(int *embix, int sourceix)
+{//======================================
 	// There were embedded commands in the text at this point
 	unsigned int word;  // bit 7=last command for this word, bits 5,6 sign, bits 0-4 command
 	unsigned int value;
 	int command;
 
 	do {
-		word = embedded_list[embix++];
+		word = embedded_list[*embix];
 		value = word >> 8;
 		command = word & 0x7f;
 
+		if(command == 0)
+			return;  // error
+
+		(*embix)++;
+
 		switch(command & 0x1f)
 		{
 		case EMBED_S:   // speed
@@ -1126,14 +1308,18 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 	PHONEME_TAB *prev_ph;
 	static int sourceix=0;
 
-#ifdef TEST_MBROLA
-	if(mbrola_name[0] != 0)
-		return(MbrolaGenerate(phoneme_list,n_ph,resume));
-#endif
+	PHONEME_DATA phdata;
+	PHONEME_DATA phdata_prev;
+	PHONEME_DATA phdata_next;
+	PHONEME_DATA phdata_tone;
+	FMT_PARAMS fmtp;
 
 	if(option_quiet)
 		return(0);
 
+	if(mbrola_name[0] != 0)
+		return(MbrolaGenerate(phoneme_list,n_ph,resume));
+
 	if(resume == 0)
 	{
 		ix = 1;
@@ -1148,6 +1334,7 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 		syllable_centre = -1;
 		last_pitch_cmd = -1;
 		memset(vowel_transition,0,sizeof(vowel_transition));
+		DoPause(0,0);    // isolate from the previous clause
 	}
 
 	while(ix < (*n_ph))
@@ -1171,13 +1358,19 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 
 		if(p->synthflags & SFLAG_EMBEDDED)
 		{
-			DoEmbedded(embedded_ix, p->sourceix);
+			DoEmbedded(&embedded_ix, p->sourceix);
 		}
 
 		if(p->newword)
 		{
-			if(translator->langopts.param[LOPT_WORD_MERGE] == 0)
+			if(((p->type == phVOWEL) && (translator->langopts.param[LOPT_WORD_MERGE] & 1)) ||
+				 (p->ph->phflags & phNOPAUSE))
+			{
+			}
+			else
+			{
 				last_frame = NULL;
+			}
 
 			sourceix = (p->sourceix & 0x7ff) + clause_start_char;
 
@@ -1210,22 +1403,39 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 
 		case phSTOP:
 			released = 0;
-			if(next->type==phVOWEL) released = 1;
-			if(next->type==phLIQUID && !next->newword) released = 1;
-
-			if(released)
-				DoSample(p->ph,next->ph,2,0,0);
+			if(next->type==phVOWEL)
+			{
+				 released = 1;
+			}
 			else
-				DoSample(p->ph,phoneme_tab[phonPAUSE],2,0,0);
+			if(!next->newword)
+			{
+				if(next->type==phLIQUID) released = 1;
+//				if(((p->ph->phflags & phPLACE) == phPLACE_blb) && (next->ph->phflags & phSIBILANT)) released = 1;
+			}
+			if(released == 0)
+				p->synthflags |= SFLAG_NEXT_PAUSE;
+
+			InterpretPhoneme(NULL, 0, p, &phdata);
+			phdata.pd_control |= pd_DONTLENGTHEN;
+			DoSample3(&phdata, 0, 0);
 			break;
 
 		case phFRICATIVE:
+			InterpretPhoneme(NULL, 0, p, &phdata);
+
 			if(p->synthflags & SFLAG_LENGTHEN)
-				DoSample(p->ph,next->ph,2,p->length,0);  // play it twice for [s:] etc.
-			DoSample(p->ph,next->ph,2,p->length,0);
+			{
+				DoSample3(&phdata, p->length, 0);  // play it twice for [s:] etc.
+			}
+			DoSample3(&phdata, p->length, 0);
 			break;
 
 		case phVSTOP:
+			ph = p->ph;
+			memset(&fmtp, 0, sizeof(fmtp));
+			fmtp.fmt_control = pd_DONTLENGTHEN;
+
 			pre_voiced = 0;
 			if(next->type==phVOWEL)
 			{
@@ -1252,11 +1462,15 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 			if((prev->type==phVOWEL) || (prev->ph->phflags & phVOWEL2))
 			{
 				// a period of voicing before the release
-				DoSpect(p->ph,phoneme_tab[phonSCHWA],next->ph,1,p,0);
+				InterpretPhoneme(NULL, 0x01, p, &phdata);
+				fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+				fmtp.fmt_amp = phdata.sound_param[pd_FMT];
+
+				DoSpect2(ph, 0, &fmtp, p, 0);
 				if(p->synthflags & SFLAG_LENGTHEN)
 				{
-					DoPause(20,0);
-					DoSpect(p->ph,phoneme_tab[phonSCHWA],next->ph,1,p,0);
+					DoPause(25,1);
+					DoSpect2(ph, 0, &fmtp, p, 0);
 				}
 			}
 			else
@@ -1271,15 +1485,24 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 			{
 				// followed by a vowel, or liquid + vowel
 				StartSyllable();
-				DoSpect(p->ph,prev->ph,next->ph,2,p,0);
 			}
 			else
 			{
-//				if((prev->type != phVOWEL) && ((prev->ph->phflags & phVOICED)==0) && ((next->ph->phflags & phVOICED)==0))
-//					DoSpect(p->ph,prev->ph,phoneme_tab[phonPAUSE_SHORT],2,p,0);
-//				else
-					DoSpect(p->ph,prev->ph,phoneme_tab[phonPAUSE],2,p,0);
-//				DoSpect(p->ph,prev->ph,next->ph,2,p,0);
+				p->synthflags |= SFLAG_NEXT_PAUSE;
+			}
+			InterpretPhoneme(NULL,0, p, &phdata);
+			fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+			fmtp.fmt_amp = phdata.sound_param[pd_FMT];
+			fmtp.wav_addr = phdata.sound_addr[pd_ADDWAV];
+			fmtp.wav_amp = phdata.sound_param[pd_ADDWAV];
+			DoSpect2(ph, 0, &fmtp, p, 0);
+
+			if((p->newword == 0) && (next2->newword == 0))
+			{
+				if(next->type == phVFRICATIVE)
+					DoPause(20,0);
+				if(next->type == phFRICATIVE)
+					DoPause(12,0);
 			}
 			break;
 
@@ -1304,22 +1527,29 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 				}
 			}
 
-			if((next->type==phVOWEL) || ((next->type==phLIQUID)) && (next->newword==0))  // ?? test 14.Aug.2007
+			if((next->type==phVOWEL) || ((next->type==phLIQUID) && (next->newword==0)))  // ?? test 14.Aug.2007
 			{
 				StartSyllable();
-				if(p->synthflags & SFLAG_LENGTHEN)
-					DoSpect(p->ph,prev->ph,next->ph,2,p,0);
-				DoSpect(p->ph,prev->ph,next->ph,2,p,0);
 			}
 			else
 			{
-				if(p->synthflags & SFLAG_LENGTHEN)
-					DoSpect(p->ph,prev->ph,phoneme_tab[phonPAUSE],2,p,0);
-				DoSpect(p->ph,prev->ph,phoneme_tab[phonPAUSE],2,p,0);
+				p->synthflags |= SFLAG_NEXT_PAUSE;
 			}
+			InterpretPhoneme(NULL,0, p, &phdata);
+			memset(&fmtp, 0, sizeof(fmtp));
+			fmtp.std_length = phdata.pd_param[i_SET_LENGTH]*2;
+			fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+			fmtp.fmt_amp = phdata.sound_param[pd_FMT];
+			fmtp.wav_addr = phdata.sound_addr[pd_ADDWAV];
+			fmtp.wav_amp = phdata.sound_param[pd_ADDWAV];
+
+			if(p->synthflags & SFLAG_LENGTHEN)
+				DoSpect2(p->ph, 0, &fmtp, p, 0);
+			DoSpect2(p->ph, 0, &fmtp, p, 0);
 			break;
 
 		case phNASAL:
+			memset(&fmtp, 0, sizeof(fmtp));
 			if(!(p->synthflags & SFLAG_SEQCONTINUE))
 			{
 				DoAmplitude(p->amp,NULL);
@@ -1331,29 +1561,32 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 				last_frame = NULL;
 			}
 
+			InterpretPhoneme(NULL,0, p, &phdata);
+			fmtp.std_length = phdata.pd_param[i_SET_LENGTH]*2;
+			fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+			fmtp.fmt_amp = phdata.sound_param[pd_FMT];
+
 			if(next->type==phVOWEL)
 			{
 				StartSyllable();
-				DoSpect(p->ph,prev->ph,next->ph,1,p,0);
+				DoSpect2(p->ph, 0, &fmtp, p, 0);
 			}
 			else
 			if(prev->type==phVOWEL && (p->synthflags & SFLAG_SEQCONTINUE))
 			{
-				DoSpect(p->ph,prev->ph,phoneme_tab[phonPAUSE],2,p,0);
+				DoSpect2(p->ph, 0, &fmtp, p, 0);
 			}
 			else
 			{
 				last_frame = NULL;  // only for nasal ?
-				if(next->type == phLIQUID)
-					DoSpect(p->ph,prev->ph,phoneme_tab[phonSONORANT],2,p,0);
-				else
-					DoSpect(p->ph,prev->ph,phoneme_tab[phonPAUSE],2,p,0);
+				DoSpect2(p->ph, 0, &fmtp, p, 0);
 				last_frame = NULL;
 			}
 
 			break;
 
 		case phLIQUID:
+			memset(&fmtp, 0, sizeof(fmtp));
 			modulation = 0;
 			if(p->ph->phflags & phTRILL)
 				modulation = 5;
@@ -1376,17 +1609,14 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 			if(next->type==phVOWEL)
 			{
 				StartSyllable();
-				DoSpect(p->ph,prev_ph,next->ph,1,p,modulation);  // (,)r
-			}
-			else
-			if(prev->type==phVOWEL && (p->synthflags & SFLAG_SEQCONTINUE))
-			{
-				DoSpect(p->ph,prev_ph,next->ph,1,p,modulation);
-			}
-			else
-			{
-				DoSpect(p->ph,prev_ph,next->ph,1,p,modulation);
 			}
+			InterpretPhoneme(NULL, 0, p, &phdata);
+			fmtp.std_length = phdata.pd_param[i_SET_LENGTH]*2;
+			fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+			fmtp.fmt_amp = phdata.sound_param[pd_FMT];
+			fmtp.wav_addr = phdata.sound_addr[pd_ADDWAV];
+			fmtp.wav_amp = phdata.sound_param[pd_ADDWAV];
+			DoSpect2(p->ph, 0, &fmtp, p, modulation);
 
 			break;
 
@@ -1394,16 +1624,48 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 			ph = p->ph;
 			stress = p->stresslevel & 0xf;
 
-			// vowel transition from the preceding phoneme
-			vowel_transition0 = vowel_transition[0];
-			vowel_transition1 = vowel_transition[1];
+			memset(&fmtp, 0, sizeof(fmtp));
+
+			InterpretPhoneme(NULL, 0, p, &phdata);
+			fmtp.std_length = phdata.pd_param[i_SET_LENGTH] * 2;
+
+			if(((fmtp.fmt_addr = phdata.sound_addr[pd_VWLSTART]) != 0) && ((phdata.pd_control & pd_FORNEXTPH) == 0))
+			{
+				// a vowel start has been specified by the Vowel program
+				fmtp.fmt_length = phdata.sound_param[pd_VWLSTART];
+			}
+			else
+			if(prev->type != phPAUSE)
+			{
+				// check the previous phoneme
+				InterpretPhoneme(NULL, 0, prev, &phdata_prev);
+				if((fmtp.fmt_addr = phdata_prev.sound_addr[pd_VWLSTART]) != 0)
+				{
+					// a vowel start has been specified by the Vowel program
+					fmtp.fmt2_lenadj = phdata_prev.sound_param[pd_VWLSTART];
+				}
+				fmtp.transition0 = phdata_prev.vowel_transition[0];
+				fmtp.transition1 = phdata_prev.vowel_transition[1];
+			}
+
+			if(fmtp.fmt_addr == 0)
+			{
+				// use the default start for this vowel
+				fmtp.use_vowelin = 1;
+				fmtp.fmt_control = 1;
+				fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+			}
+
+			fmtp.fmt_amp = phdata.sound_param[pd_FMT];
 
 			pitch_env = envelope_data[p->env];
 			amp_env = NULL;
 			if(p->tone_ph != 0)
 			{
-				pitch_env = LookupEnvelope(phoneme_tab[p->tone_ph]->spect);
-				amp_env = LookupEnvelope(phoneme_tab[p->tone_ph]->after);
+				InterpretPhoneme2(p->tone_ph, &phdata_tone);
+				pitch_env = GetEnvelope(phdata_tone.pitch_env);
+				if(phdata_tone.amp_env > 0)
+					amp_env = GetEnvelope(phdata_tone.amp_env);
 			}
 
 			StartSyllable();
@@ -1419,13 +1681,13 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 			{
 				DoAmplitude(p->amp,amp_env);
 				DoPitch(pitch_env,p->pitch1,p->pitch2);  // don't use prevocalic rising tone
-				DoSpect(ph,prev->ph,next->ph,1,p,modulation);
+				DoSpect2(ph, 1, &fmtp, p, modulation);
 			}
 			else
 			if(prev->type==phLIQUID || prev->type==phNASAL)
 			{
 				DoAmplitude(p->amp,amp_env);
-				DoSpect(ph,prev->ph,next->ph,1,p,modulation);   // continue with pre-vocalic rising tone
+				DoSpect2(ph, 1, &fmtp, p, modulation);  // continue with pre-vocalic rising tone
 				DoPitch(pitch_env,p->pitch1,p->pitch2);
 			}
 			else
@@ -1436,7 +1698,7 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 					DoPitch(pitch_env,p->pitch1,p->pitch2);
 				}
 
-				DoSpect(ph,prev->ph,next->ph,1,p,modulation);
+				DoSpect2(ph, 1, &fmtp, p, modulation);
 			}
 
 			if(option_phoneme_events)
@@ -1444,9 +1706,33 @@ int Generate(PHONEME_LIST *phoneme_list, int *n_ph, int resume)
 				DoMarker(espeakEVENT_PHONEME, sourceix, 0, p->ph->mnemonic);
 			}
 
-			DoSpect(p->ph,prev->ph,next->ph,2,p,modulation);
+			fmtp.fmt_addr = phdata.sound_addr[pd_FMT];
+			fmtp.fmt_amp = phdata.sound_param[pd_FMT];
+			fmtp.transition0 = 0;
+			fmtp.transition1 = 0;
+
+			if((fmtp.fmt2_addr = phdata.sound_addr[pd_VWLEND]) != 0)
+			{
+				fmtp.fmt2_lenadj = phdata.sound_param[pd_VWLEND];
+			}
+			else
+			if(next->type != phPAUSE)
+			{
+				fmtp.fmt2_lenadj = 0;
+				InterpretPhoneme(NULL, 0, next, &phdata_next);
+
+				fmtp.use_vowelin = 1;
+				fmtp.transition0 = phdata_next.vowel_transition[2];  // always do vowel_transition, even if ph_VWLEND ??  consider [N]
+				fmtp.transition1 = phdata_next.vowel_transition[3];
+
+				if((fmtp.fmt2_addr = phdata_next.sound_addr[pd_VWLEND]) != 0)
+				{
+					fmtp.fmt2_lenadj = phdata_next.sound_param[pd_VWLEND];
+				}
+			}
+
+			DoSpect2(ph, 2, &fmtp, p, modulation);
 
-			memset(vowel_transition,0,sizeof(vowel_transition));
 			break;
 		}
 		ix++;
@@ -1599,20 +1885,23 @@ int SpeakNextClause(FILE *f_in, const void *text_in, int control)
 	CalcPitches(translator, clause_tone);
 	CalcLengths(translator);
 
-	GetTranslatedPhonemeString(translator->phon_out,sizeof(translator->phon_out));
-	if(option_phonemes > 0)
+	if((option_phonemes > 0) || (phoneme_callback != NULL))
 	{
-		fprintf(f_trans,"%s\n",translator->phon_out);
-
-		if(!iswalpha(0x010d))
+		GetTranslatedPhonemeString(translator->phon_out,sizeof(translator->phon_out));
+		if(option_phonemes > 0)
 		{
-			// check that c-caron is recognized as an alphabetic character
-			fprintf(stderr,"Warning: Accented letters are not recognized, eg: U+010D\nSet LC_CTYPE to a UTF-8 locale\n");
+			fprintf(f_trans,"%s\n",translator->phon_out);
+	
+			if(!iswalpha(0x010d))
+			{
+				// check that c-caron is recognized as an alphabetic character
+				fprintf(stderr,"Warning: Accented letters are not recognized, eg: U+010D\nSet LC_CTYPE to a UTF-8 locale\n");
+			}
+		}
+		if(phoneme_callback != NULL)
+		{
+			phoneme_callback(translator->phon_out);
 		}
-	}
-	if(phoneme_callback != NULL)
-	{
-		phoneme_callback(translator->phon_out);
 	}
 
 
@@ -1622,20 +1911,6 @@ int SpeakNextClause(FILE *f_in, const void *text_in, int control)
 		return(1);
 	}
 
-	if(mbrola_name[0] != 0)
-	{
-#ifdef USE_MBROLA_LIB
-		MbrolaTranslate(phoneme_list,n_phoneme_list,NULL);
-#else
-		{
-			FILE *f_mbrola;
-			if((f_mbrola = f_trans) == stderr)
-				f_mbrola = stdout;
-			MbrolaTranslate(phoneme_list,n_phoneme_list,f_mbrola);
-		}
-#endif
-	}
-
 	Generate(phoneme_list,&n_phoneme_list,0);
 	WavegenOpenSound();
 
